"""
Logging module

"""
import logging

DEFAULT_LEVEL = 'INFO'

DEFAULT_LOG_NAME = 'disco_stu.log'

#LOG_LEVELS = {DEBUG: 10, STDINFO: 21, INFO: 20}

# Formatting
MAX_LEVEL_NAME_LENGTH = 7
MAX_WIDTH = 80

LOG_FORMAT = ('%(asctime)s %(name)s: %(levelname)-{0}s - '
              '%(message)s'.format(MAX_LEVEL_NAME_LENGTH))
CONSOLE_FORMAT = '    %(message)s'

def get_logger(name):
    """
    Convenience function to customize and configure logger

    Parameters
    ----------

    name : str
     Name to instantiate log file with.

    Returns
    -------
    : :py:class:`~logging.Logger`

    """
    log = logging.getLogger(name)
    try:
        assert log.root.handlers
    except AssertionError:
        log = configure_log()

    return log

def configure_log(logfile=DEFAULT_LOG_NAME, log_level=DEFAULT_LEVEL):
    """
    Configure a custom logger. If log file is None a default log file name is
    used. Appends to log file if it exists otherwise creates it.

    Parameters
    ----------

    logger : :py:class:`~logging.Logger`

    logfile : str
       File to log to

    log_level : str
        {0}, default: {1}

    Returns
    -------
    
    : :py:class:`~logging.Logger`
        Logger configured and customized.

    """
    logging_level = getattr(logging, log_level)

    # Create logger object with name of module
    log = logging.getLogger('')

    rootlog = log
    for handler in rootlog.handlers:
        if isinstance(handler, logging.StreamHandler):
            handler.flush()
        rootlog.removeHandler(handler)

    # Log will always record debug statements; but only log_level will go to
    # console
    logging.basicConfig(level=logging.DEBUG,
                        format=LOG_FORMAT,
                        datefmt='%Y-%m-%d %H:%M:%S', filename=logfile,
                        filemode='a')

    # Define a Handler which writes logging_level messages or higher to the
    # sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging_level)

    # Set different format for STDERR
    formatter = logging.Formatter(CONSOLE_FORMAT)
    console.setFormatter(formatter)
    console.setLevel(log_level.upper())

    # add the handler to the root logger
    log.addHandler(console)

    return log