import numpy as np
from astropy import log as astropy_log
from astropy.nddata import NDData
from astropy.io import fits
from astropy.table import Table
from astropy.modeling import models
from astropy.wcs import WCS, FITSFixedWarning
import copy, os, warnings
import math
from datetime import datetime

from .nduncertainty import VarUncertainty
from ..lookups import gsaoi_static_distortion_info as gsdi
from ..lookups import general_parameters as params
from ..transform import CallableWCS, Transform, make_tnx_model

from ..bookkeeping.log_utilities import get_logger, MAX_WIDTH

log = get_logger(__name__)

# Suppress those annoying INFO messages about overwriting WCS or meta
astropy_log.setLevel('WARNING')

class BartData(object):
    """
    A class for storing astronomical images as a series of NDData objects.
    The I/O here is generic, but the other attached methods are for GSAOI
    only, since that's what we working with.
    
    This is not a full implementation, but only what is required here.
    """
    def __init__(self, source=None, identifier=None, arrays=None):
        self._phu = None
        self._data_objects = {}

        # Does this work?
        if isinstance(source, fits.Header):
            self._phu = source
            if arrays:
                for count, array in enumerate(arrays, start=1):
                    if isinstance(array, NDData):
                        self[count] = array
        else:
            
            if os.path.isfile(source):
                self.readFITS(source)
                if identifier is None:
                    identifier = source.split('/')[-1].split('.')[0]
            else:
                raise IOError("Can't read {}".format(source))
        self._identifier = identifier
            
    def __getitem__(self, extver):
        return self._data_objects[extver]
    
    def __setitem__(self, extver, array):
        assert isinstance(array, NDData)
        array.meta.update({'extver': extver})
        self._data_objects.update({extver: array})

    def __len__(self):
        return len(self._data_objects)
    
    def __iter__(self):
        self._index = 1
        return self
    
    # Python 3 compatible
    def __next__(self):
        if self._index > len(self):
            raise StopIteration
        self._index += 1
        return self._data_objects[self._index-1]
    
    # Python 3 compatible
    def __next__(self):
        if self._index > len(self):
            raise StopIteration
        self._index += 1
        return self._data_objects[self._index-1]
    
    # Python 2 compatible
    next = __next__
        
    @property
    def phu(self):
        return self._phu
    
    @property
    def identifier(self):
        return self._identifier

    @identifier.setter
    def identifier(self, value):
        self._identifier = str(value)
    #-------------------------------------------------------------------------            
    def readFITS(self, source):
        """
        Read a FITS file into a BartData instance
        
        Parameters
        ----------
        source: string
          Name of the input file on disk
        """
        self._phu = None
        hdulist = fits.open(source)
        
        # Populate dicts so we can handle any order of extensions
        dict_data = {}
        dict_dq = {}
        dict_var = {}
        dict_meta = {}
        next_extver = 1
        for hdu in hdulist:
            if hdu.name == 'PRIMARY':
                self._phu = hdu.header
                if hdu.data is None:
                    continue
            if not hdu.name:
                hdu.name = 'SCI'
            if 'EXTVER' in hdu.header:
                if hdu.header['EXTVER'] == -1:
                    hdu.header['EXTVER'] = next_extver
                    next_extver += 1
            else:
                hdu.header['EXTVER'] = next_extver
                next_extver += 1    
            extver = int(hdu.header['EXTVER'])
            
            # Should be able to deal with non-Gemini data. A PrimaryHDU must
            # have a data attribute to have got to this point.
            if hdu.name=='SCI' or hdu.name=='PRIMARY':
                dict_data.update({extver: hdu.data})
                if extver not in dict_meta:
                    dict_meta.update({extver: {}})
                dict_meta[extver].update({'hdu': hdu.header})
                # Set DQ and VAR to none to avoid having to test later
                # Only if not already present; can be overwritten later
                if extver not in dict_dq:
                    dict_dq.update({extver: None})
                if extver not in dict_var:
                    dict_var.update({extver: None})
            elif hdu.name == 'DQ':
                dict_dq.update({extver: hdu.data.astype(params.DQ_DATATYPE)})
            elif hdu.name == 'VAR':
                dict_var.update({extver: VarUncertainty(hdu.data)})
            else:
                key = hdu.header['EXTNAME']
                if isinstance(hdu, fits.ImageHDU):
                    value = hdu.data
                elif isinstance(hdu, fits.BinTableHDU):
                    value = Table(hdu.data, meta=dict(hdu.header))
                # TODO Think about things without EXTVER
                if extver > 0:
                    if extver not in dict_meta:
                        dict_meta.update({extver: {}})
                    dict_meta[extver].update({key: value})   
        hdulist.close()
        
        # Make the BartData object
        for extver in dict_data:
            dict_meta[extver].update({'extver': extver})
            try:
                with warnings.catch_warnings():
                    warnings.filterwarnings('ignore',
                                            category=FITSFixedWarning)   
                    wcs = Transform(CallableWCS(WCS(dict_meta[extver]['hdu'])))
            except Exception as err:
                wcs = None
            self._data_objects.update({extver: NDData(dict_data[extver],
                    meta=dict_meta[extver], mask=dict_dq[extver],
                    uncertainty=dict_var[extver], wcs=wcs)})
            # Have to set this manually until NDData fix is released
            if dict_var[extver] is not None:
                self._data_objects[extver].uncertainty.parent_nddata = \
                    self._data_objects[extver]
    #-------------------------------------------------------------------------            
    def writeFITS(self, filename=None, overwrite=True):
        """
        Write the BartData instance as a FITS file
        
        Parameters
        ----------
        filename: string
          Name of the outputfile (if None, use identifier)
          
        overwrite: bool
          Allow overwriting of an existing file?
        """
        if filename is None:
            if self._identifier is None:
                filename = 'default.fits'
            else:
                filename = self._identifier+'.fits'
        
        hdulist = fits.HDUList([fits.PrimaryHDU(header=self._phu)])
        hdulist[0].header['EXTEND'] = True
        # TODO Write out extensions in sensible order
        for data_object in self:
            hdr = data_object.meta['hdu']
            hdr.update({'EXTVER': data_object.meta['extver']})
            hdulist.append(fits.ImageHDU(data_object.data,
                                         header=hdr, name='SCI'))
            if data_object.uncertainty is not None:
                hdulist.append(fits.ImageHDU(data_object.uncertainty.array,
                                             header=hdr, name='VAR'))
            if data_object.mask is not None:
                hdulist.append(fits.ImageHDU(data_object.mask,
                                             header=hdr, name='DQ'))
            for key, value in data_object.meta.items():
                if isinstance(value, np.ndarray):
                    hdulist.append(fits.ImageHDU(data=value,
                                                header=hdr, name=key))
                if isinstance(value, Table):
                    hdulist.append(fits.BinTableHDU(data=value.as_array(),
                        header=fits.Header(value.meta.items()), name=key))
        try:
            hdulist.writeto(filename, overwrite=overwrite)
        except AttributeError:
            hdulist.writeto(filename, clobber=overwrite)

    #-------------------------------------------------------------------------
    def add_timestamp(self, update):
        """
        Add a FITS Header keyword as a timestamp
        Input: keyword name (string) or tuple of keyword name and comment
        """
        if isinstance(update, tuple):
            key, comment = update
        else:
            key, comment = (update, None)

        time_stamp = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
        self._phu.set(key, time_stamp, comment)
    #-------------------------------------------------------------------------            
    def attach_static_transform(self):
        """
        Attach a Transform object as the WCS to each array
        """
        # This is pointless unless the instrument is GSAOI
        if self._phu.get('INSTRUME') != 'GSAOI':
            return
        
        # Calculate the PA transformation (same for all arrays)
        pa_corr_transform = pa_correction(self)
        
        for array in self:
            # Set up a new (empty) transformation
            wcs_transform = Transform()

            # Correct for ROI
            wcs_transform.update(roi_correction(array))
            
            # Attach transform
            wcs_transform.update(default_static_model(array))
            
            # Scale from arcseconds to degrees
            wcs_transform.update((models.Scale(1./3600) &
                            models.Scale(1./3600)).rename('ArcsecToDegrees'))
            
            # Correct for PA
            wcs_transform.update(pa_corr_transform)
            
            # Correct for pointing
            wcs_transform.update(pointing_correction(array, self._phu))
                        
            # Can't overwrite wcs of existing NDData object so make new one
            extver = array.meta['extver']
            self._data_objects.update({extver:
                                       NDData(array, wcs=wcs_transform)})
    #-------------------------------------------------------------------------                        
    def detect_sources(self, detect_function, overwrite=False, **kwargs):
        """
        Attach a Table of SExtractor output to each array
        Inputs: detect_function = function to perform the source detection
                overwrite = ignore any existing OBJCAT?
                **kwargs = additional parameters supplied to detect_function
        """
        total_sources = 0
        for array in self:
            extver = array.meta['extver']
            if 'OBJCAT' in array.meta and not overwrite:
                if np.all([c in array.meta['OBJCAT'].columns
                           for c in ['X_IMAGE','Y_IMAGE']]):
                    array_sources = len(array.meta['OBJCAT'])
                    log.debug("Found OBJCAT in {}:{} with {} sources".
                        format(self._identifier, extver, array_sources))
                    total_sources += array_sources
                    array.meta['OBJCAT']['ARRAY'] = extver
                    continue
                else:
                    log.warning("{}:{} has an OBJCAT but cannot find X_IMAGE,"
                        "Y_IMAGE columns".format(self._identifier, extver))
            objcat = detect_function(array, **kwargs)
            array_sources = len(objcat)
            if array_sources>0:
                objcat['ARRAY'] = array.meta['extver']
                array.meta.update({'OBJCAT': objcat})
            log.debug("Detected {} sources in {}:{}".
                      format(array_sources, self._identifier, extver))
            total_sources += array_sources
        log.info("Found {} sources in {}".format(total_sources,
                                                    self._identifier))
    #-------------------------------------------------------------------------            
    def measure_sky(self, measure_sky_function, **kwargs):
        """
        Measure the sky level for each detector and store it in the meta
        """
        for array in self:
            sky_level = measure_sky_function(array, **kwargs)
            log.debug("Measured sky level of {0:.2f} in {1}:{2}".
                    format(sky_level, self._identifier, array.meta['extver']))
            array.meta.update({'skylevel': sky_level})
########################### End of BartData Class ############################

def roi_correction(array):
    """
    Calculate the required shift if the array is not the full ROI
    Input: array = NDData instance for the array
    Output: Transform object (if needed) or None (if not)
    """
    [x1,x2,y1,y2] = [int(a)
        for b in array.meta['hdu']['CCDSEC'].strip('[]').split(':')
            for a in b.split(",")]
    roi = [y1-1,y2, x1-1,x2]
    shifts = [this_value-default_value for this_value, default_value
              in zip(roi, gsdi.TRIMMED_FULL_FRAME_ROI)]

    transform = (models.Shift(shifts[2]) &
                 models.Shift(shifts[0])).rename('ROICOrrection') \
        if (shifts[0] or shifts[2]) else None

    return transform
#-----------------------------------------------------------------------------
def pa_correction(data):
    """
    Calculate the required shift if the array is rotated. The correction
      is the same for all arrays
    Input: data = BartData instance
    Output: Transform object (if needed) or None (if not)
    """
    pa = data._phu['PA']
    # IAA not included (why not? what if it changes?)
    # pa -= data._phu['IAA']
    
    # Don't bother if the shift is <1 pixel across the entire instrument
    # This is the correction we're applying to de-rotate the instrument
    # so the sign is negative compared to the PHU keyword
    transform = models.Rotation2D(-pa).rename('PACorrection') \
        if abs(pa)>0.01 else None

    return transform
#-----------------------------------------------------------------------------
def default_static_model(array, date=None):
    """
    Calculate the appropriate static distortion model for this array
    using the coefficients in the lookup table
    Input: data = NDData instance for the array
    Output: list of models to accomplish the mapping
    """
    
    extver = array.meta['extver']
    
    # This is here in case the correction changes with time
    date_dict = gsdi.STATIC_CORRECTIONS_DATE_MAPPING
    for (start_date, end_date) in date_dict.keys():
        return_key = (start_date, end_date)
        if end_date is None:
            if start_date is None or date is None:
                break
            elif date >= start_date:
                break
        elif date >= start_date and date < end_date:
            break
    else:
        raise IOError("Cannot determine which static correction to use")
    
    # Retrieve the static_correction_dictionaries
    top_level_dict = date_dict[return_key]
    full_forward_dict = getattr(gsdi, top_level_dict[gsdi.MODEL])
    full_reverse_dict = getattr(gsdi, top_level_dict[gsdi.INVERSE])
    cor_offset = getattr(gsdi, top_level_dict[gsdi.CENTRE_OF_ROTATION_OFFSET])

    # Correction for trimming (4 pixels) plus 1-indexing. Lazily hardcoded.
    indexing_transform = (models.Shift(cor_offset['x']+5) &
                models.Shift(cor_offset['y']+5)).rename('IndexingTransform')

    # Pix2TNX projection transform
    forward_model = []
    for model in full_forward_dict[gsdi.CORRECTION_ORDER]:
        model_dict = full_forward_dict[model][gsdi.ARRAY_NAMES["A"+str(extver)]]
        model_dict.update({'model': full_forward_dict[model][gsdi.MODEL]})
        forward_model.append(model_dict)
    reverse_model = []
    for model in full_reverse_dict[gsdi.CORRECTION_ORDER]:
        model_dict = full_reverse_dict[model][gsdi.ARRAY_NAMES["A"+str(extver)]]
        model_dict.update({'model': full_reverse_dict[model][gsdi.MODEL]})
        reverse_model.append(model_dict)
    pix2sky_tnx_model = make_tnx_model(forward_model, reverse_model)
    
    return [indexing_transform, pix2sky_tnx_model]
#-----------------------------------------------------------------------------
def pointing_correction(array, phu):
    """
    Inputs: array = NDDdata instance for the array
            phu = image primary header unit (has pointing keywords)
    Output: model to accomplish the mapping
    """
    ra = phu['RA']
    dec = phu['DEC']
    
    # Correct for RA and DEC offsets
    dec += phu['DECOFFSE'] / 3600
    ra += phu['RAOFFSET'] / (3600 * math.cos(math.radians(dec)))
        
    pointing_transform = (models.Pix2Sky_TAN() |
        models.RotateNative2Celestial(ra, dec,
            180)).rename('Pointing Correction')
    
    return pointing_transform
