from astropy.nddata import NDData
from astropy.modeling import models
from collections import OrderedDict

from .data_handling.multinddata import BartData
from .transform import calculate_output_projection
from .matching import align_images
from .stack import stack_images
from .projection import get_output_projection, apply_projection, project_image
from .bookkeeping.log_utilities import get_logger
from .lookups import general_parameters as params
from .source_detection.source_detection import sextractor_detection
from .source_detection.sextractoreti import SExtractorETI
from .lookups import source_detection_info as sdi

log = get_logger(__name__)

def disco(infiles,
           output_identifier,
           objmatch_info,
           refmatch_info,
           static_transform=True,
           ignore_objcat=False,
           refcat=None,
           skysub=True,
           stack=True,
           align=True,
           combine_type=params.COMBINE_TYPE,
           reference_image=None,
           donor_wcs=None,
           final_pa=None,
           pixel_scale=None,
           parallel=True,
           clean_iter=0,
           clean_footprint=None):
    """
    Runs the show.
    
    Parameters
    ----------
    infiles: list
      Filenames of files to be stacked
      
    output_identifier: string
      Filename of output stacked image
      
    objmatch_info: MatchInfo namedtuple
      offset_radius: float
        radius (in arcseconds) for systemic alignment shift search
      match_radius: float
        radius (in arcseconds) for closest neighbour match
      min_matches: int (or None)
        minimum number of matches needed to perform alignment
      degree: int
        degree of polynomial for alignment transform
        
    refmatch_info: MatchInfo namedtuple
      as objmatch_info, but for OBJCAT-to-REFCAT matching
      
    static_transform: bool
      Apply GSAOI static transform before catalogue matching?
      
    ignore_objcat: bool
      Run SExtractor to detect sources even if an image has an OBJCAT extension?
      
    refcat: Table
      Reference catalogue for absolute astrometry (RA,DEC columns only, as
      it's already been processed)
      
    skysub: bool
      Measure and subtract the background from each array?
      
    stack: bool
      Stack images?
      
    align: bool
      Align images? May not want to if using previously reprojected images
      
    combine_type: string
      Type of combine operation to perform (mean, average, median)

    reference_image: string
      Name of reference image (otherwise, use first image in list)
      
    donor_wcs: string
      Name of image from which to take WCS
      
    final_pa: float
      Position angle for +ve y-axis of final image (overrides PA of first
      image in stack)
      
    pixel_scale: float
      Pixel scale (in arcseconds) of final image
      
    parallel: bool
      Undertake transformation of individual arrays in parallel? (for
      multi-array instruments like GSAOI)
      
    clean_iter: int
      Maximum number of iterations to undertake when cleaning bad pixels
    """

    # This will be a dict of useful information to allow the image alignment
    # and projection to be calculated
    datasets = OrderedDict()

    if reference_image is not None and reference_image not in infiles:
        log.warning("Reference image {} not in input list. Using first "
                    "file.".format(reference_image))
        reference_image = None
    if reference_image is None:
        reference_image = infiles[0]
    reference_image = reference_image.replace('.fits', '')

    if align:
        # Initialize ETI for SExtractor
        eti=SExtractorETI(sdi.DEFAULT_SOURCE_DETECTION_FILES)

        log.info("Reading input files and running source detection...")
        reference = None
        for infile in infiles:
            dataset = BartData(infile)
            dataset.detect_sources(sextractor_detection,
                                   overwrite=ignore_objcat, eti=eti)
    
            if params.PACKAGE_MARKER[0] in dataset.phu:
                log.warning("{} has already been processed by {}; will not"
                    " apply transform".format(dataset.identifier,
                                        params.PACKAGE_NAME.title()))
            else:
                dataset.attach_static_transform()

            # Store everything in a dict so we can work without the pixels
            dataset_dict = {}
            for array in dataset:
                value = {'shape': array.data.shape, 'wcs': array.wcs}
                if 'OBJCAT' in array.meta:
                    value.update({'OBJCAT': array.meta['OBJCAT']})
                dataset_dict.update({array.meta['extver']: value})
            datasets.update({dataset.identifier: dataset_dict})
            
            if reference_image == dataset.identifier:
                reference = dataset.identifier
                reference_phu = dataset.phu

        log.info("Matching sources...")
        datasets = align_images(datasets, reference, objmatch_info,
                                refmatch_info, models.Polynomial2D, refcat)
        
        output_shape = None
        if donor_wcs is not None:
            log.info("Reading output WCS from {}...".format(donor_wcs))
            output_projection, output_shape = get_output_projection(donor_wcs)
        
        # Not an "else" in case that failed...
        if output_shape is None:
            log.info("Calculating the output projection...")
            output_projection, output_shape = calculate_output_projection(
                            datasets, reference_phu, projection=params.PROJECTION,
                            final_pa=final_pa, pixel_scale=pixel_scale)

        log.info("Output image will be {0} pixels".format(" x ".join(
                        [repr(npix) for npix in output_shape[::-1]])))

        output_template = params.PROJ_ID
        for infile in infiles:
            project_image(infile, output_template, datasets, output_projection,
                          output_shape, skysub=skysub, clean_iter=clean_iter,
                          clean_footprint=clean_footprint, parallel=parallel)

        # Create new input list for stacking            
        infiles = []
        for identifier in datasets:
            infiles.append(output_template.format(identifier)+'.fits')

    # Stack images. Keep the BartData stuff here so stack_images only has
    # to deal with NDData instances
    if stack:
        log.info("Stacking images...")
        datasets = [BartData(infile) for infile in infiles]
        
        # These images only have a single EXTVER=1 array
        output_image = stack_images([dataset[1] for dataset in datasets],
                                    combine_type=combine_type)
        output_dataset = BartData(datasets[0].phu, identifier=output_identifier,
                                  arrays=[output_image])
        output_dataset.add_timestamp(params.PACKAGE_MARKER)
        log.info("Writing output image {}.fits...".format(output_identifier))
        output_dataset.writeFITS(overwrite=True)
