import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.modeling import models, fitting
from astropy.nddata import NDData
import math

from .utils import merge_objcats
from .bookkeeping.log_utilities import get_logger

log = get_logger(__name__)

def align_images(datasets, reference, objmatch_info, refmatch_info, model_type,
                 refcat=None, grow=True):
    """
    Return a list of BartData instances with modified WCS objects derived by
    matching and aligning to either a reference catalogue (if supplied)
    or the first image in the input list
    Inputs: datasets = list of BartData instances
            objmatch_info = namedtuple of parameters for OBJCAT-to-OBJCAT matching
            refmatch_info = namedtuple of parameters for OBJCAT-to-REFCAT matching
            model_type = type of Model to fit
            refcat = Table of reference source positions (RA, DEC)
            grow = add unmatched objects from images to source catalog?
    Outputs: list of BartData instances
    """
    
    # Note: it might appear to make sense to match to the REFCAT first.
    # Then all images can have that transformation applied at the same
    # time as the transformation to match to the first image. The reason
    # for not doing this is that it would be good to use all the entire
    # field coverage for REFCAT matching, and not just the first image's
    # FOV. Such a method requires that all input images be on a single
    # coordinate frame (i.e., that defined by the first image) first, and
    # so the order used here will be necessary.

    # Use the WCS transforms to find RA and Dec of objects
    for dataset in datasets.values():
        for array in dataset.values():
            if 'OBJCAT' in array:
                objcat = array['OBJCAT']
                ra, dec = array['wcs'](objcat['X_IMAGE'].astype(float)-1,
                                       objcat['Y_IMAGE'].astype(float)-1)
                objcat['RA'] = ra
                objcat['RA'].unit = u.degree
                objcat['DEC'] = dec
                objcat['DEC'].unit = u.degree
    
    # Better matching occurs if the images are all aligned to one another
    # before thinking about the reference catalog. This makes sense given
    # there will be many more objects in the images than the REFCAT.
    log.info("Aligning all images to {0}...".format(reference))
    ref_objcat = merge_objcats(datasets[reference])
    for identifier, dataset in datasets.items():
        if identifier != reference:
            full_objcat = merge_objcats(dataset)
            alignment_transform, num_matches = _calculate_alignment_transform(
                    ref_objcat, full_objcat,
                    objmatch_info.offset_radius, objmatch_info.match_radius, 
                    model_type, objmatch_info.degree)
            log.info("Matched {0} sources between {1} and {2}".
                    format(num_matches, identifier, reference))
            if objmatch_info.min_matches is not None and num_matches < objmatch_info.min_matches:
                alignment_transform = None
            if alignment_transform is None:
                log.warning("Too few matches! No alignment will be performed!")
            # But still need to propagate OBJCATs back to individual arrays
            # since they have mRA,mDEC columns now
            for extver,array in dataset.items():
                # Extract the OBJCAT rows corresponding to this array.
                if 'OBJCAT' in array:
                    array.update({'OBJCAT': 
                                  full_objcat[full_objcat['ARRAY']==extver]})

                # Update WCS to include transformation to first image's WCS
                if alignment_transform is not None:
                    wcs_transform = array['wcs']
                    wcs_transform.update(alignment_transform)
                    array.update({'wcs': wcs_transform})
                    # Add in unmatched sources to the source catalog, *after*
                    # transforming them to first image's WCS!
                    if grow and 'OBJCAT' in array:
                        new_objcat = array['OBJCAT'][np.where(
                                            array['OBJCAT']['mRA']<0)]
                        new_objcat.remove_columns(['mRA','mDEC'])
                        tra, tdec = alignment_transform(new_objcat['RA'],
                                                        new_objcat['DEC'])
                        new_objcat['RA'] = tra
                        new_objcat['DEC'] = tdec
                        for row in new_objcat:
                            log.debug('Adding source {} {}'.format(row['RA'], row['DEC']))
                            ref_objcat.add_row(row)
        
    # Now align images to reference catalog, if provided.
    # TODO? Average object positions across datasets.
    if refcat:
        refcat_alignment_transform, num_matches = _calculate_alignment_transform(
                        refcat, ref_objcat,
                        refmatch_info.offset_radius, refmatch_info.match_radius,
                        model_type, refmatch_info.degree)        
                
        log.info("Matched {0} sources between reference catalog and {1}".
                    format(num_matches, reference))
        if refmatch_info.min_matches is not None and num_matches < refmatch_info.min_matches:
            refcat_alignment_transform = None
        if refcat_alignment_transform is None:
            log.warning("Too few matches! No alignment will be performed!")
        else:
            # Recreate the first image's individual array's OBJCATs
            for extver,array in datasets[reference].items():
                array.update({'OBJCAT': 
                              ref_objcat[ref_objcat['ARRAY']==extver]})
    else:
        refcat_alignment_transform = None
        # Need to add dummy matched coordinate columns for consistency with
        # the other tables and subsequent code
        for array in datasets[reference].values():
            if 'OBJCAT' in array:
                array['OBJCAT']['mRA'] = -99.0
                array['OBJCAT']['mDEC'] = -99.0

    log.info("Updating object catalogs...")            
    # Now apply the REFCAT matching transform (if it exists), and update the OBJCAT
    for identifier, dataset in datasets.items():
        resid_ra = []
        resid_dec = []
        for array in dataset.values():
            if refcat_alignment_transform is not None:
                wcs_transform = array['wcs']
                wcs_transform.update(refcat_alignment_transform)
                array.update({'wcs': wcs_transform})

            # Now update the OBJCAT will all these transformations
            # The full transformation to final celestial coordinates
            if 'OBJCAT' in array:
                objcat = array['OBJCAT']
                objcat['tRA'], objcat['tDEC'] = array['wcs'](
                            objcat['X_IMAGE'].astype(float)-1,
                            objcat['Y_IMAGE'].astype(float)-1)
                # Update matched coordinates for all images but the first
                # (remember, these are the coordinates in the first image's
                # static transformation, which has been tweaked)
                if refcat_alignment_transform is not None and identifier!=reference:
                    mra, mdec = refcat_alignment_transform(objcat['mRA'], objcat['mDEC'])
                    objcat['mRA'] = np.where(objcat['mRA']>-90, mra, -99.0)
                    objcat['mDEC'] = np.where(objcat['mDEC']>-90, mdec, -99.0)
                    
                # Residuals (in arcseconds)
                dec_factor = math.cos(math.radians(np.mean(objcat['tDEC'])))
                rra = (objcat['mRA'] - objcat['tRA']) * 3600 * dec_factor
                rdec = (objcat['mDEC'] - objcat['tDEC']) * 3600
                resid_ra.extend(rra[objcat['mDEC']>-90])
                resid_dec.extend(rdec[objcat['mDEC']>-90])
                objcat['rRA'] = np.where(objcat['mDEC']<-90, 0, rra)
                objcat['rDEC'] = np.where(objcat['mDEC']<-90, 0, rdec)
        if resid_ra and resid_dec:
            log.debug('Mean and rms residuals for {}: ({:.4f},{:.4f}) ({:.4f},{:.4f})'.
                  format(identifier,np.mean(resid_ra),np.mean(resid_dec),
                         np.std(resid_ra),np.std(resid_dec)))

    return datasets
#-----------------------------------------------------------------------------
def _calculate_alignment_transform(refcat, objcat, offset_search_radius,
                                   match_radius, model_type, degree):
    """
    Matches two catalogues and calculates the mapping needed to align them
    
    Inputs: refcat = Table with RA,DEC of reference sources
            objcat = Table with RA,DEC of input sources from frame to align
            offset_search_radius = radius (in arcsec) to determine the
                                   systemic offset between coord frames
            match_radius = radius (in arcsec) for the object-to-object
                           matching, after applying the systemic offset
            model_type = type of Model to fit
            degree = degree of this model
    Outputs: alignment_transform = Model instance describing the mapping
             num_matches = number of sources matched
    """
        
    xin, yin, xout, yout = match_catalogues(refcat, objcat,
                                offset_search_radius=offset_search_radius,
                                match_radius=match_radius)
    num_matches = len(xin)
    
    # No point trying to compute a transform if there aren't enough
    # matched sources for it to be constrained; downgrade degree of fit
    use_degree = degree
    while (len(model_type(degree=use_degree).parameters)>num_matches and
           use_degree>1):
        use_degree -= 1

    if use_degree>0 and len(model_type(degree=use_degree).parameters)<=num_matches:
        if use_degree<degree:
            log.debug("Downgrading fit to degree {} since only {} matches".
                      format(use_degree,num_matches))      
        # Perform the fitting
        m_init = model_type(degree=use_degree)
        f = fitting.LinearLSQFitter()
        ra_transform = f(m_init, xin, yin, xout)
        dec_transform = f(m_init, xin, yin, yout)
        ra_transform_inverse = f(m_init, xout, yout, xin)
        dec_transform_inverse = f(m_init, xout, yout, yin)
        
        # Need to duplicate inputs since both the RA and DEC transforms
        # require both coordinates as inputs
        alignment_transform = models.Mapping((0,1,0,1)) | (ra_transform &
                                                           dec_transform)
        alignment_transform.inverse = models.Mapping((0,1,0,1)) | \
            (ra_transform_inverse & dec_transform_inverse)
        log.debug('RA transform: '+repr(ra_transform))
        log.debug('DEC transform:'+repr(dec_transform))
    else:
        alignment_transform = None
    
    return alignment_transform, num_matches
#-----------------------------------------------------------------------------
def match_catalogues(refcat, objcat, offset_search_radius, match_radius):
    """
    Calculate a Model that maps the RA and dec from objcat to refcat
    Inputs: refcat = Table containing RA and DEC columns to be mapped to
            objcat = Table containing RA and DEC columns
            offset_search_radius = radius (in arcsec) to determine the
                                   systemic offset between coord frames
            match_radius = radius (in arcsec) for the object-to-object
                           matching, after applying the systemic offset
    Output: tuple of 4 arrays of x,y coords from objcat and x,y coords
            from refcat that match
    """
    
    # Get coordinates from catalogues
    # TODO probably want to abstract this so it can do any coordinate matching
    xin = objcat['RA'].data
    yin = objcat['DEC'].data
    xout = refcat['RA'].data
    yout = refcat['DEC'].data
    
    # There's a fair chance that OBJCAT will have far more sources than REFCAT
    # which could cause bright REFCAT sources to match to faint OBJCAT ones
    # so we should shrink the OBJCAT to a sensible size
    if len(objcat) > 2*len(refcat) and 'FLUX_AUTO' in objcat.colnames:
        log.debug("Shrinking OBJCAT from {} sources to {}".
                  format(len(objcat),2*len(refcat)))
        indices = np.argsort(objcat['FLUX_AUTO'].data.astype(float))[-2*len(refcat):]
    else:
        indices = np.arange(len(objcat), dtype=int)

    # Before matching sources between the catalogues, we need to determine the
    # systematic offset. This should be accounted for by RAOFFSET, DECOFFSE
    # but there will be imperfections at the <1-arcsec level, which is enough
    # to affect the matching.
    # This is a Cartesian alignment task, so we need to make (RA,DEC) roughly
    # Cartesian by applying the usual cos(DEC) factor to the RA coordinates.
    dec_factor = math.cos(math.radians(np.mean(yout)))
    # Don't try this if we're so close to the pole that things will go weird
    if dec_factor > 0.01:
        dx, dy = _get_offset(xin[indices]*dec_factor, yin[indices],
                             xout*dec_factor, yout,
                             offset_search_radius/3600)
        log.debug("Systemic offsets measured as ({0:6.3f},{1:6.3f})".
                    format(dx*3600,dy*3600))
    else:
        dx, dy = 0, 0
        match_radius = offset_search_radius
    
    refcoords = SkyCoord(xout, yout, unit=u.degree)
    catcoords = SkyCoord(xin[indices]+dx/dec_factor, yin[indices]+dy,
                         unit=u.degree)
    idx, sep2d, dist3d = catcoords.match_to_catalog_sky(refcoords)
    
    # Use coordinates only of sources that matched
    matched = np.where(sep2d<match_radius*u.arcsecond)[0]
    objcat['mRA'] = -99.0
    objcat['mDEC'] = -99.0
    for match in matched:
        objcat['mRA'][indices[match]] = xout[idx[match]]
        objcat['mDEC'][indices[match]] = yout[idx[match]]

    xin = xin[indices[matched]]
    yin = yin[indices[matched]]
    xout = xout[idx[matched]]
    yout = yout[idx[matched]]
    
    return (xin, yin, xout, yout)
#-----------------------------------------------------------------------------
def _get_offset(xin,yin, xref,yref, delta, precision=0.02/3600):
    """
    Return offset to shift (xin,yin) to (xref,yref)
    Inputs: xin,yin = coordinates of sources in input image
            xref,yref = coordinates of sources in reference image
            delta = maximum search radius
            precision = required accuracy
    Outputs: xoffset,yoffset = shift required
    """
    
    # Exit if either list is empty
    if len(xin)*len(xref) == 0:
        return (0,0)
    
    # The method here is to do cross-correlation by making a synthetic image
    # of the reference source locations and moving the input source locations
    # about to find the best match. We need to have a suitable pixel grid for
    # the synthetic image, and map all coordinates onto this grid
    x0 = np.min(np.concatenate((xin,xref)))
    y0 = np.min(np.concatenate((yin,yref)))
    xoffsets = []
    yoffsets = []

    fak = 20
    hw = fak * 2
    # A scale factor; each pixel in the synthetic image corresponds to this
    # amount in the actual coordinates. We're going to do this iteratively in
    # ever decreasing circles until we have an accurate offset
    scale = delta / fak * 5
    while scale>precision or not xoffsets:
        scale *= 0.2
        xx = (xin-x0 + np.sum(xoffsets))/scale + hw
        yy = (yin-y0 + np.sum(yoffsets))/scale + hw
        sx = (xref-x0)/scale + hw
        sy = (yref-y0)/scale + hw
        
        sigmasq = (0.25*hw)**2
        # Make a "landscape" of Gaussian "mountains" onto which we're
        # going to cross-correlate the REFCAT sources
        landscape = np.zeros((int(np.max(sy)+hw),int(np.max(sx)+hw)))
        lysize, lxsize = landscape.shape
        xgrid, ygrid = np.mgrid[0:hw*2+1,0:hw*2+1]
        rsq = (ygrid-hw)**2 + (xgrid-hw)**2
        mountain = np.exp(-0.5*rsq/sigmasq)
        for x, y in zip(sx, sy):
            mx1, mx2, my1, my2 = 0, hw*2+1, 0, hw*2+1
            lx1, lx2 = int(x)-hw, int(x)+hw+1
            ly1, ly2 = int(y)-hw, int(y)+hw+1
            # None of this should happen but leave it in for safety
            if lx2<0 or lx1>=lxsize or ly2<0 or ly1>=lysize:
                continue
            if lx1 < 0:
                mx1 -= lx1
                lx1 = 0
            if lx2 > lxsize:
                mx2 -= (lx2-lxsize)
                lx2 = lxsize
            if ly1 < 0:
                my1 -= ly1
                ly1 = 0
            if ly2 > lysize:
                my2 -= (ly2-lysize)
                ly2 = lysize
            try:
                landscape[ly1:ly2,lx1:lx2] += mountain[my1:my2,mx1:mx2]
            except ValueError as e:
                # And this certainly shouldn't happen!
                log.debug('{} {} {}'.format(y, x, landscape.shape))
    
        # Undertake a grid search to find the best offset. Simply move the input
        # coordinates about and sum the heights of the landscape at the points
        # to which these move.
        dax = []; day = []; height = []
        for dx in np.arange(-fak,fak+1):
            for dy in np.arange(-fak,fak+1):
                # For each shift, sum the landscape pixel values at all shifted
                # coordinates -- remember to test whether they're in the landscape
                new_xx = (xx+dx+0.5).astype(int)
                new_yy = (yy+dy+0.5).astype(int)
                indices = np.all((new_xx>=0,new_xx<lxsize,
                            new_yy>=0,new_yy<lysize),axis=0)
                dax.append(dx)
                day.append(dy)
                height.append(np.sum(landscape[new_yy[indices],new_xx[indices]]))
                
        xoffset, yoffset = dax[np.argmax(height)]*scale, day[np.argmax(height)]*scale
        xoffsets.append(xoffset)
        yoffsets.append(yoffset)
        log.debug("Shift of ({},{})".format(xoffset,yoffset))
        
    # Add up all the offsets we've measured so far
    return np.sum(xoffsets), np.sum(yoffsets)
