import numpy as np
from astropy.nddata import NDData
from astropy.coordinates import SkyCoord
from astropy import units as u
from astropy.modeling import models, projections
from scipy.ndimage import geometric_transform, filters
from copy import deepcopy
import multiprocessing as multi
import math

from .data_handling.nduncertainty import VarUncertainty
from .data_handling.multinddata import BartData
from .lookups import general_parameters as params
from .transform import Transform, CallableWCS, GeoMapping
from .utils import measure_bg_from_image, merge_objcats

from .bookkeeping.log_utilities import get_logger

log = get_logger(__name__)

#-----------------------------------------------------------------------------
def get_output_projection(donor_image):
    """
    Derive an output projection from the size and WCS of another image
    Output must be compatible with that of calculate_output_projection
    Input: donor_image = filename of donor_image
    Outputs: output_map = list of models to make the output projection
             output_shape = tuple giving dimensions of output image
    Returns None, None if there's a problem
    """
    
    # Try to open the image
    try:
        donor = BartData(donor_image)
    except IOError:
        log.warning("Cannot open {}".format(donor_image))
        return None, None
    
    output_map = None
    output_shape = None
    for array in donor:
        wcs = array.wcs
        if wcs is not None:
            if isinstance(wcs.map(), CallableWCS):
                output_map = wcs.map().inverse
                output_shape = array.data.shape
                break

    return output_map, output_shape
#-----------------------------------------------------------------------------
def project_image(infile, output_template, datasets, output_projection,
                  output_shape, skysub=True, clean_iter=0,
                  clean_footprint=None, parallel=True):
    """
    Self-contained routine to read in a file, project it, and write it out.
    This is designed to be parallelized.
    Inputs: infile = filename of input image
            output_template = string describing how to make output image name
            datasets = dict containing WCS and OBJCAT information
            output_projection = transformation from (RA,DEC) to output pixels
            output_shape = dimensions of output image
            skysub = perform sky subtraction?
            clean_iter = max number of cleaning iterations to perform
            clean_footprint = footprint to use for median filter
            parallel = transform arrays simultaneously?
    """

    dataset = BartData(infile)
    identifier = dataset.identifier
    if skysub:
        dataset.measure_sky(measure_bg_from_image)
        
    # Update WCS of the arrays from calculated transform
    for array in dataset:
        extver = array.meta['extver']
        dataset[extver] = NDData(array, wcs=datasets[identifier][extver]['wcs'])
        try:
            array.meta.update({'OBJCAT': datasets[identifier][extver]['OBJCAT']})
        except KeyError:
            pass

    output_image = apply_projection(dataset, output_projection, output_shape,
                        clean_iter=clean_iter, clean_footprint=clean_footprint,
                        parallel=parallel)
    
    output_identifier = output_template.format(identifier)
    new_dataset = BartData(dataset.phu, identifier=output_identifier,
                           arrays=[output_image])
    
    log.info("Writing {0} to disk".format(output_identifier))
    new_dataset.add_timestamp(params.PACKAGE_MARKER)
    new_dataset.writeFITS(overwrite=True)
#-----------------------------------------------------------------------------
def _apply_geometric_transform(input, transform, output, output_shape,
                               reverse=False):
    """
    None-returning function to apply geometric transform, so it can be used
    by multiprocessing
    Inputs: input = dict of input arrays
            transform = Transform object to apply
            output = DictProxy to hold the transformed output
            output_shape = tuple indicating the shape of these outputs
            reverse = apply transform in reverse?
    """
    
    # Set up the mapping from output coords to input coords
    keys = input.keys()
    input_shape = input[list(keys)[0]].shape
    
    # Ensure all inputs and outputs have the same shape
    assert np.all([input[key].shape == input_shape for key in keys]), \
        "Not all input arrays have the same shape!"

    # Save memory by having these as 4-byte floats
    xin = np.empty(output_shape, dtype=np.float32)
    yin = np.empty(output_shape, dtype=np.float32)
    
    # Memory-saving framework. Would like to have some standard module to
    # check available memory (it's not worth packing psutil here).
    nsub = output_shape[0]*output_shape[1] // 25000000 + 1
    row_limits = np.linspace(0, output_shape[0], nsub+1, dtype=np.int16)
    for i in range(nsub):
        xgrid, ygrid = np.meshgrid(np.arange(output_shape[1]),
                                   np.arange(row_limits[i],row_limits[i+1]))
        xx, yy = transform(xgrid, ygrid, reverse=True)
        xin[row_limits[i]:row_limits[i+1]] = xx
        yin[row_limits[i]:row_limits[i+1]] = yy
    xin = np.where(xin > input_shape[1]-1, -1, xin)
    yin = np.where(yin > input_shape[0]-1, -1, yin)
    map = GeoMapping(xin, yin)
    
    # Run the transform and store the result
    for key in keys:
        cval = 16 if key[1]=='DQ' else 0
        out_array = geometric_transform(input[key].astype(np.float32), map, output_shape,
                                        order=params.INTERPOLATION_ORDER, cval=cval)
        output[key] = out_array
#-----------------------------------------------------------------------------
def apply_projection(input, output_projection, output_shape, clean_iter=0,
                     clean_footprint=None, parallel=True):
    """
    Reproject an input image via its own WCS and a defined output projection
    Inputs: input = BartData instance to reproject
            output_projection = transformation from (RA,dec) to output plane
            output_shape = shape of output image
            clean_iter = max number of cleaning iterations to perform
            clean_footprint = footprint to use for median filter
            parallel = apply multiprocessing (uses extra memory, has OSX issues)
    Output: New NDData instance
    """
    
    # Calculate pixel size of input array to allow flux conservation
    # Since flatfielding makes the input image effectively in surface
    # brightness units, we need to scale all images by the same amount,
    # even though the effective pixel sizes are different on each array
    pixscale_in = np.mean([_calculate_pixel_scale(0.5*array.data.shape[1],
            0.5*array.data.shape[0], array.wcs) for array in input])
    pixscale_out = _calculate_pixel_scale(0.5*output_shape[1],
                    0.5*output_shape[0],
                    Transform(output_projection).map(reverse=True))
    area_scale = (pixscale_out / pixscale_in)**2
    log.debug("{}: Pixel scale input, output, scale factor {}, {}, {}".
              format(input.identifier, pixscale_in, pixscale_out, area_scale))

    # Set up for multiprocessing
    if parallel:
        processes = []
        manager = multi.Manager()
        output_arrays = manager.dict()
    else:
        output_arrays = {}

    # There's no point appending VAR and DQ unless all arrays have them
    # so these booleans indicate whether to use the VAR and DQ planes
    dq_exists = np.all([array.mask is not None for array in input])
    var_exists = np.all([array.uncertainty is not None for array in input])
    output_data = np.zeros(output_shape, dtype=np.float32)
    if var_exists:
        output_uncertainty = np.zeros(output_shape, dtype=np.float32)
    if dq_exists:
        output_dq = np.full(output_shape, params.NO_DATA, dtype=np.float32)

    log.info("Applying transform for {}".format(input.identifier))
    for array in input:
        # Make complete transform from input pixels to output pixels
        ysize_in, xsize_in = array.data.shape
        full_transform = deepcopy(array.wcs)
        full_transform.update(output_projection)
        
        # In order to speed things up, determine the rectangle circumscribing
        # the region of the output image into which this array will go, and
        # only map that
        x_corners = [0, xsize_in-1, xsize_in-1, 0]
        y_corners = [0, 0, ysize_in-1, ysize_in-1]
        xout, yout = full_transform(x_corners, y_corners)
        x1 = int(np.floor(np.min(xout)))
        x2 = int(np.ceil(np.max(xout))) + 1
        y1 = int(np.floor(np.min(yout)))
        y2 = int(np.ceil(np.max(yout))) + 1
        log.debug("{}:{} maps to output image x1={} x2={} y1={} y2={}".
                  format(input.identifier, array.meta['extver'], x1,x2,y1,y2))
        if x1 < 0:
            x1 = 0
        if x2 > output_shape[1]:
            x2 = output_shape[1]
        if y1 < 0:
            y1 = 0
        if y2 > output_shape[0]:
            y2 = output_shape[0]
        
        # So this image goes into [y1:y2, x1:x2]
        full_transform.update((models.Shift(-x1) & models.Shift(-y1)).
                              rename('Region offset'))
        
        # We only want actual *bad* pixels to be interpolated over, not
        # saturated or NO_DATA pixels, hence bitmask=1
        if clean_iter>0:
            array = _clean_data(array, 1, clean_iter, clean_footprint)
        
        # Outputs will be returned as items in a dict, keyed by a double:
        # a tuple indicating their location in the full output image
        # and a string indicating the SCI, VAR, DQ
        # TODO All a bit hardcoded; should do better
        key = (x1,x2,y1,y2)
        array_output_shape = (y2-y1,x2-x1)
        input_arrays = {}
        # If sky-subtraction is on, we'll ignore the SKYMIDPT keyword
        if 'skylevel' in array.meta:
            skylevel = array.meta['skylevel']
        elif 'SKYMIDPT' in array.meta['hdu']:
            skylevel = array.meta['hdu']['SKYMIDPT'] / array.meta['hdu']['GAIN']
        else:
            skylevel = None

        if skylevel:
            input_arrays[key, 'SCI'] = array.data - skylevel
            log.debug("Subtracting sky level of {} from {}:{}".
                    format(skylevel, input.identifier, array.meta['extver']))
        else:
            input_arrays[key, 'SCI'] = array.data
        
        # It's too time-consuming to transform each bad pixel bit plane,
        # so set pixels to simply either "good" or "bad"
        if dq_exists:
            input_arrays[key, 'DQ'] = np.where(array.mask &
                                               params.DQ_BITMASK, 1, 0)
        # Simply interpolate the VAR array
        if var_exists:
            input_arrays[key, 'VAR'] = array.uncertainty.array

        # Set up process is parallelized; just do it(TM) if not
        if parallel:
            processes.append(multi.Process(target=_apply_geometric_transform,
                args=(input_arrays, full_transform, output_arrays, array_output_shape),
                kwargs={'reverse': True}))
        else:
            _apply_geometric_transform(input_arrays, full_transform, output_arrays,
                                       array_output_shape, reverse=True)

        if 'OBJCAT' in array.meta:        
            objcat = array.meta['OBJCAT']
            tx, ty = Transform(output_projection)(objcat['tRA'], objcat['tDEC'])
            # Store where this object is in the output image
            objcat['tX_IMAGE'] = tx+1 # 1-indexed
            objcat['tY_IMAGE'] = ty+1
            ox, oy = full_transform(tx-x1,ty-y1, reverse=True)
            # And which pixel in the original image that comes from
            # This is a good distortion-studying test because these coordinates
            # should be almost identical to the original X_IMAGE,Y_IMAGE
            objcat['oX_IMAGE'] = ox+1
            objcat['oY_IMAGE'] = oy+1

    if parallel:        
        # Start the processes...     
        for p in processes:
            p.start()
        # ...and wait for them all to finish        
        for p in processes:
            p.join()

    # Now plonk the individual transformed arrays into the full image
    for key in output_arrays.keys():
        (x1,x2,y1,y2), type = key
        log.debug("Placing a {} array in ({},{},{},{})".format(type,x1,x2,y1,y2))
        if type == 'SCI':
            output_data[y1:y2, x1:x2] += output_arrays[key] * area_scale
        elif type == 'VAR':
            # Interpolation can make values<0 if the VAR is rapidly changing
            # Need to do something about this; set VAR=0 will effectively
            # ignore them in the stack
            output_uncertainty[y1:y2, x1:x2] += np.where(output_arrays[key]>0,
                    output_arrays[key]*np.sqrt(area_scale),0)
        elif type == 'DQ':
            # If we're about to stack a NO_DATA pixel, ignore it and keep
            # whatever is already there (which might also be NO_DATA)
            output_dq[y1:y2, x1:x2] = np.where(output_arrays[key]==params.NO_DATA,
                            output_dq[y1:y2, x1:x2], output_arrays[key])

    # Tidy up the DQ: Only values between 0 and MIN_DQ_BADNESS are good
    if dq_exists:
        output_dq = np.where(np.fabs(output_dq) < params.MIN_DQ_BADNESS,
                             0, output_dq)
        output_dq = np.where(np.logical_and(output_dq !=0,
                                            output_dq != params.NO_DATA), 1, output_dq)

    # Output image HDU will be made from the input array headers.
    # Need to fix the WCS for the new projection
    hdu = _update_header([array.meta['hdu'] for array in input],
                         output_projection, id=input.identifier)
    
    # Add the sky level back in if fl_autosky=yes during gareduce, which
    # creates the SKYMIDPT keyword (in electrons)
    if 'SKYMIDPT' in hdu:
        output_data += hdu['SKYMIDPT'] / hdu['GAIN'] * area_scale
    
    full_objcat = merge_objcats(input)

    output_image = NDData(data=output_data, meta={'hdu': hdu, 'OBJCAT': full_objcat})
    if dq_exists:
        output_image.mask = output_dq.astype(params.DQ_DATATYPE)
    if var_exists:
        output_image.uncertainty = VarUncertainty(output_uncertainty)
        output_image.uncertainty.parent_nddata = output_image
                
    return output_image
#-----------------------------------------------------------------------------
def _clean_data(array, bitmask=65535, max_iter=2,
                footprint=np.array([[1,1,1],[1,0,1],[1,1,1]])):
    """
    Attempt to fix bad pixels in the SCI plane using neighbouring pixel values
    This should avoid any ringing that might arise from interpolation
    Inputs: array = NDData instance of array to fix
            bitmask = pixels to flag as bad
            max_iter = maximum number of iterations 
            footprint = median filter
    Output: "fixed" version of the input (only .data is changed)
    """
    # No-op quickly if there's no DQ plane
    if array.mask is None or bitmask==0 or max_iter<1:
        return array
    mask = (array.mask & bitmask)>0

    iter = 0
    filtered_data = array.data
    while (iter<max_iter and np.any(mask)):
        median_data = filters.median_filter(filtered_data, footprint=footprint)
        filtered_data = np.where(mask, median_data, filtered_data)
        iter += 1
        # Don't waste time here if we're leaving anyway
        # This version of the mask should indicate which pixels are still bad
        # because half (or more) the pixels used to calculate the median were bad
        if iter < max_iter:
            mask = filters.median_filter(mask, footprint=footprint)

    # Of course, we have to make the entire NDData object again    
    return NDData(filtered_data, mask=array.mask, uncertainty=array.uncertainty,
                  wcs=array.wcs, meta=array.meta)
    
#-----------------------------------------------------------------------------
def _calculate_pixel_scale(xc, yc, model):
    """
    Return pixel scale (arcsec/pixel) for a given transformation
    Inputs: xc, yc = pixel location about which to determine pixel scale
            model = Model describing transformation from (x,y) to (RA,dec)
    Output: pixel scale (arcsec/pixel)
    """
    xin = [xc-0.5, xc+0.5, xc, xc]
    yin = [yc, yc, yc-0.5, yc+0.5]
    xout, yout = model(xin, yin)
    sky_coords = [SkyCoord(ra*u.deg, dec*u.deg) for ra,dec in zip(xout, yout)]
    cdelt1 = sky_coords[0].separation(sky_coords[1]).arcsecond
    cdelt2 = sky_coords[2].separation(sky_coords[3]).arcsecond
    return np.sqrt(cdelt1*cdelt2)
#-----------------------------------------------------------------------------
def _update_header(headers, output_projection, id=''):
    """
    Fix the FITS WCS header keywords to be consistent with the data, and delete
    extraneous keywords
    Inputs: header = dict-like object
            output_projection = list of models describing the output projection
            id = image identifier (for logging purposes)
    Output: new version of header with updated WCS
    
    These (identical) calculations are done for each output image, to
    enable generality in the future. We have defined the output projection
    so could set these keywords from that.
    """
    
    newkeys = {}
    delkeys = ['CCDNAME','CCDSIZE','CCDSUM','CCDSEC','DETSEC','DATASEC',
               'ARRAYID','FRMNAME','FRAMEID','TRIMSEC']
    [delkeys.extend(['AMPSEC{}'.format(array),'AMPNAME{}'.format(array)])
     for array in range(1,5)]
    
    # If we've been donated a WCS, use the WCS method to populate the header
    if isinstance(output_projection, CallableWCS):
        newkeys = output_projection.wcs.to_header()
        # WCS.to_header() uses PCi_j keywords, so add CDi_j keywords
        if 'CD1_1' not in newkeys:
            [newkeys.update({'CD{}_{}'.format(i,j):
                    newkeys['CDELT{}'.format(i)]*newkeys['PC{}_{}'.format(i,j)]
                    if 'PC{}_{}'.format(i,j) in newkeys else 0})
             for i in [1,2] for j in [1,2]]
    else:
        ra = None
        dec = None
        projection_type = None
        # Learn about the output projection model and base RA and DEC
        for model in output_projection:
            if isinstance(model, models.RotateCelestial2Native):
                # Parameter names changed after Astropy v1.0
                try:
                    ra = model.lon.value
                    dec = model.lat.value
                except AttributeError:
                    ra = model.phi.value
                    dec = model.theta.value
            if isinstance(model, projections.Zenithal):
                # Astropy v>1.0 renames projections so find 3-letter version
                for code in projections.projcodes:
                    try:
                        if model.__class__ == getattr(projections, 'Sky2Pix_'+code):
                            projection_type = code
                    except AttributeError:
                        pass
                
        assert ra and dec and projection_type, "Cannot produce FITS header"
    
        newkeys['CRVAL1'] = ra
        newkeys['CRVAL2'] = dec
        newkeys['CTYPE1'] = 'RA---'+projection_type
        newkeys['CTYPE2'] = 'DEC--'+projection_type
    
        # Ideally, use the output projection to work out the pixel location, but
        # this has led to some issues because of floating point inaccuracies.
        # Instead, extract the values of the shift from the final element of the
        # output projection list
        xc = output_projection[-1].offset_0.value
        yc = output_projection[-1].offset_1.value
        newkeys['CRPIX1'] = xc + 1 # FITS coords are 1-indexed
        newkeys['CRPIX2'] = yc + 1
    
        # ...and the pixel transformation matrix around this location
        xin = [xc-0.5, xc+0.5, xc, xc]
        yin = [yc, yc, yc-0.5, yc+0.5]
        xout, yout = Transform(output_projection)(xin, yin, reverse=True)
        newkeys['CD1_1'] = (xout[1]-xout[0]) * math.cos(math.radians(dec))
        newkeys['CD1_2'] = (xout[3]-xout[2]) * math.cos(math.radians(dec))
        newkeys['CD2_1'] = yout[1]-yout[0]
        newkeys['CD2_2'] = yout[3]-yout[2]

    # Use mean of the GAIN and RDNOISE in output header
    for key in ['GAIN','RDNOISE']:
        if np.all([key in h for h in headers]):
            newkeys.update({key: float('{:.3f}'.format(np.mean([h[key]
                            for h in headers])))})
    
    header = headers[0]
    header.update(newkeys)
    [log.debug("Updating {} header keyword {} with value {}".
                  format(id, key,newkeys[key])) for key in newkeys]
        
    [header.remove(key) for key in delkeys if key in header]
    
    return header
