from __future__ import absolute_import

import re
import subprocess

from .eti import ExternalTaskInterface as ETI
from .sextractoretifile import SExtractorETIFile
from .sextractoretiparam import SExtractorETIParam

from ..lookups import source_detection_info as sdi

#log = logutils.get_logger(__name__)
__SYS_COMMAND__ = "sex"
__MIN_VERSION__ = [2, 8, 6]
__VERSION_COMMAND__ = [__SYS_COMMAND__, "--version"]
__REGEXP_GROUP_NAME__ = "v1"
__VERSION_REGEXP__ = ''.join(["^.*version (?P<", __REGEXP_GROUP_NAME__,
                              r">[\d+\.]+) .*$"])

class SExtractorETI(ETI):
    """This class coordinates the ETI as is relates to sextractor"""
    def __init__(self, parameters=None):
        self._set_name(__name__)
        super(SExtractorETI, self).__init__()
        self._version = None
        self._check_version()
        self.add_param(SExtractorETIParam(parameters))

    def _check_version(self):
        """Only check version once"""
        if self._version is None:
            self._version = \
                self._test_version(command=__VERSION_COMMAND__,
                                   minimum_version=__MIN_VERSION__)

    def _version_regexp(self):
        # Compile a regular expression for matching the version
        # number from sextractor output
        return (re.compile(__VERSION_REGEXP__), __REGEXP_GROUP_NAME__)


    def _test_version(self, command, minimum_version=list):
        """
        Returns True if command is runnable and version is
        minimum_version or later. Expects version is a dot seperated list of
        numbers, with the same number of values as minimum_version.

        """
        # Get acceptable version from global variable
        (stdoutdata, stderrdata) = self._execute(command,
                                                   return_output=True)

        version_regexp, group_names = self._version_regexp()
        if isinstance(group_names, list):
            if len(group_names) != len(minimum_version):
                errmsg = ("Length of regexp groups {0} is not equal to length"
                          "of minimum_version {1}".format(len(group_names),
                                                          len(minimum_version)))
                raise IOError(errmsg)
            version = ".".join([version_regexp.match(stdoutdata).group(name)
                                for name in group_names])
        else:
            version = version_regexp.match(stdoutdata).group(group_names)

        if version is None:
            errmsg = ("Unable to determine {0} version".format(self._taskname))
            raise Exception(errmsg)

        match_values = [int(i) for i in version.split(".")]
        if match_values < minimum_version:
            errmsg = ("Version {0} of {1} is required. Version "
                      "{2} installed".format('.'.join(minimum_version),
                                             self._taskname,
                                             '.'.join(match_values)))
            raise Exception(errmsg)

        return True
    
    def run(self, input):
        # Need to start with an empty input list
        self.file_objs = []
        if isinstance(input, list):
            [self.add_file(SExtractorETIFile(x)) for x in input]
        else:
            self.add_file(SExtractorETIFile(input))
        self.prepare()
        self.execute()
        objcats = self.recover()
        self.clean()
        # Don't return a 1-element list 
        return objcats[0] if len(objcats)==1 else objcats

    def execute(self):
        param_dict = {}
        for obj in self.param_objs:
            param_dict.update(obj.get_parameter())
        cmd = [__SYS_COMMAND__]
        
        # Add the config (.sex) file
        cmd.extend(['-c', param_dict.pop(sdi.CONFIGURATION)])
        
        # Add all the other command-line arguments
        for parameter, value in param_dict.items():
            cmd.extend(['-'+parameter, value])
        
        # Run SExtractor for each input file
        for file_obj in self.file_objs:
            if file_obj._dq_image is not None:
                cmd.extend(['-FLAG_IMAGE', file_obj._dq_image])
            cmd.extend(['-CATALOG_NAME', file_obj._catalog_file])
            cmd.append(file_obj._sci_image)
            self._execute(cmd, return_output=False)
            
    def recover(self):
        for par in self.param_objs:
            par.recover()
        # Return a list of OBJCATs
        return [fil.recover() for fil in self.file_objs]

    def _execute(self, command=list, return_output=False):
        """
        Call subprocess to execute shell 'command'. Tests the excuted task's
        returncode: raises if non-zero. Raises on OSError.

        If exceptions are raise the recover method is invoked.

        @command: list: Shell command to execute

        @return_output: bool: If True returns tuple (stdoutdata, stderrdata)

        """
        stdoutdata = None
        stderrdata = None
        try:
            # Use universal_newlines = True as sextractor outputs non-ascii
            # characters
            pipe_out = subprocess.Popen(command,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE,
                                        universal_newlines=True)

            # Get the output from sextractor and set returncode (communicate
            # does this) then check it
            (stdoutdata, stderrdata) = pipe_out.communicate()
            if pipe_out.returncode != 0:
                errmsg = ("SExtractor returned an error:\n"
                          "{0}{1}".format(stdoutdata, stderrdata))

                raise Exception(errmsg)
        except OSError:
            self.clean()
            raise
        else:
            # TODO Search for WARNINGS in stderrdata an print
            if return_output:
                return (stdoutdata, stderrdata)