"""
Source detection

Parameters for source detection - other than filenames and file types are
stored in the 'default.*' files in the source detection folder in the lookups
directory.

Contents:

    Functions:
        detect_sources

"""
import numpy as np

from astropy.io import fits
from astropy.table import Table
from ..bookkeeping.log_utilities import get_logger
from ..lookups import source_detection_info as sdi
from datetime import datetime

log = get_logger(__name__)

CONTAMINATION_FACTOR = 0.25

# Required columns for source detection filtering
REQUIRED_SEXTRACT_COLUMNS = ['FLAGS', 'NIMAFLAGS_ISO', 'ISOAREAF_IMAGE',
                             'ISOAREA_IMAGE', 'FWHM_IMAGE', 'ELLIPTICITY']

def filter_sources(table):
    # Check that required columns are in place
    missing_columns = [column for column in REQUIRED_SEXTRACT_COLUMNS
                       if column not in table.colnames]

    if missing_columns:
        err_msg = ("Missing columns from source detection output: "
                   "{0}".format(missing_columns))
        raise KeyError(err_msg)

    # Flag the bad sources for exclusion

    # Sextractor flags
    flag_flags = table['FLAGS'] >= 8

    # ISOAREAF_IMAGE number of pixels that weer detected, NIMAFLAGS_ISO are the
    # number of data quality pixels that were > 0 that were included in the
    # detected pixels
    contaminated_flags = table['NIMAFLAGS_ISO'] > \
            CONTAMINATION_FACTOR * table['ISOAREAF_IMAGE']

    # Number of pixels used to calculate the ISO information
    area_flags = table['ISOAREA_IMAGE'] <= 10

    fwhm_flags = table['FWHM_IMAGE'] < 1.5
    narrow_flags = table['B_IMAGE'] < 1.1
    ellipticity_flags = table['ELLIPTICITY'] >= 0.5

    # TODO Make this into a OR of a list
    rows_to_exclude = np.where(flag_flags | fwhm_flags | ellipticity_flags |
                               area_flags | narrow_flags | contaminated_flags)

    table.remove_rows(rows_to_exclude)
    
def sextractor_detection(array, eti=None, clip=True, clip_func=filter_sources):
    """
    Call SExtractor to detect objects on this array and return an
    astropy.Table object of the catalogue
    """

    if clip:
        assert callable(clip_func)

    # Set up SExtractor
    file_dict = sdi.DEFAULT_SOURCE_DETECTION_FILES

    # Use the SExtractor ETI to perform source detection
    objcat = eti.run(array)
    log.debug("Returned from ETI with {} sources".format(len(objcat)))
    
    if clip:
        clip_func(objcat)
        
    return objcat