import numpy as np
from astropy.nddata import NDData
from functools import reduce

from .data_handling.nduncertainty import VarUncertainty
from .lookups.general_parameters import DQ_DATATYPE

def _check_images(images):
    """Check all images can be stacked, i.e., are the same size"""
    ok = True
    for axis in range(2):
        ok |= len(set(image.data.shape[axis] for image in images)) <= 1
    return ok
#-----------------------------------------------------------------------------
def none_free(list):
    """Is this list free of None elements?"""
    return np.all([obj is not None for obj in list])
#-----------------------------------------------------------------------------
def stack_images(images, combine_type='average', nsub=20):
    """
    Stack a bunch of NDData instances. VAR and DQ may exist.
    Return a single NDData instance
    Inputs: images = list of NDData objects
            combine_type = np.ma function to combine
            nsub = number of subslices for stacking; there is a huge memory
                   penalty for stacking in one go, but no apparent speed
                   penalty for splitting up the input, so make this large!
    Output: single combined NDData instance, with first image's meta
    """

    if not _check_images(images):
        raise Error("Cannot stack images of different sizes")

    data_list = [image.data for image in images]
    mask_list = [image.mask for image in images]
    var_list = [None if not image.uncertainty
                else image.uncertainty.array for image in images]
    use_mask = none_free(mask_list)
    use_weights = none_free(var_list) and combine_type!='median'

    # Since np.dstack has to make a copy of all the data, this doubles the
    # memory usage. Instead, stack the image in segments. This could be
    # parallelized but it's pretty quick as a sequential task so why bother?
    output_shape = data_list[0].shape
    output_data = np.empty(output_shape, dtype=np.float32)
    output_mask = np.empty(output_shape, dtype=DQ_DATATYPE) if use_mask \
        else None
    output_var = np.empty_like(output_data) if use_weights else None

    # Only want the HDU, not the OBJCAT. Make some header keyword changes too.
    output_meta = {'hdu': images[0].meta['hdu']}
    ncombine = len(images)
    output_meta['hdu'].update({'NCOMBINE': (ncombine, 'Number of images combined')})
    gain_eff = np.sum(image.meta['hdu'].get('GAIN', 0) for image in images)
    if gain_eff > 0:
        output_meta['hdu'].update({
                'GAIN': (gain_eff, 'Effective gain (e-/ADU)'),
                'GAINORIG': (gain_eff/ncombine, 'Original gain (e-/ADU')})

    # Set up the regions... groups of rows
    row_limits = np.linspace(0, output_shape[0], nsub+1, dtype=np.int32)
    for i in range(nsub):
        substack(data_list, mask_list, var_list,
                 output_data, output_mask, output_var,
                 row_limits[i:i+2], combine_type)

    # Make the output NDData object    
    output_image = NDData(data=output_data, meta=output_meta)
    if output_mask is not None:
        output_image.mask = output_mask
    if output_var is not None:
        output_image.uncertainty = VarUncertainty(output_var)
        output_image.uncertainty.parent_nddata = output_image
    
    return output_image
#-----------------------------------------------------------------------------
def substack(data_list, mask_list, var_list, 
             output_data, output_mask, output_var,
             row_limits, combine_type):
    """
    Stack a bit of the images
    Inputs: data_list = list of data np.ndarray instances
            mask_list = list of the masks to apply
            var_list = list of the variance planes
            output_data = full array for the output
            output_mask = full array for the mask (None if no masking)
            output_var = full array for the variance (None if no weighting)
            row_limits = 2-element list indicating range to work on
            combine_type = np.ma function to combine
        -- output_data, output_mask, output_var are modified by this function
    """

    # In numpy v1.10, this is basically just np.stack()
    data_cube = np.empty((len(data_list), row_limits[1]-row_limits[0],
                          data_list[0].shape[1]), dtype=np.float32)
    for i, obj in enumerate(data_list):
        data_cube[i] = obj[row_limits[0]:row_limits[1]]

    # Do the same for the mask        
    if output_mask is not None:
        mask_cube = np.empty_like(data_cube, dtype=DQ_DATATYPE)
        for i, obj in enumerate(mask_list):
            mask_cube[i] = obj[row_limits[0]:row_limits[1]]
    else:
        mask_cube = None

    # More complex here. For average, we want the weights (1/variance)
    # but for mean, we want the variances          
    kwargs = {}
    if output_var is not None:
        aux_cube = np.empty_like(data_cube)
        for i, obj in enumerate(var_list):
            var_plane = obj[row_limits[0]:row_limits[1]]
            if combine_type == 'average':
                with np.errstate(divide='ignore', over='ignore'):
                    weight_plane = 1.0 / var_plane
                    weight_plane[var_plane==0] = 0.
                aux_cube[i] = np.copy(weight_plane)
                kwargs = {'weights': aux_cube, 'returned': True}
            elif combine_type == 'mean':
                aux_cube[i] = np.copy(var_plane)

    # Stack!
    operator = getattr(np.ma, combine_type)
    result = operator(np.ma.array(data_cube, mask=mask_cube), axis=0, **kwargs)
    
    # Interpret the output
    if isinstance(result, tuple):
        output_data[row_limits[0]:row_limits[1]] = result[0]
        tmp_mask = np.where(result[0].mask, 1, 0)
        # Get the uncertainty as an unmasked array
        with np.errstate(divide='ignore'):
            tmp_var = np.array(1.0 / result[1])
            tmp_var[result[1]==0] = 0.
    else:
        output_data[row_limits[0]:row_limits[1]] = result
        tmp_mask = np.where(result.mask, 1, 0)

    # What we can't trivially do is determine the variance from a
    # straight mean, given the masking, require var=sum(var)/n^2
    # or var=mean(var)/n, where n is the number of good pixels used
    if combine_type == 'mean' and output_var is not None:
        tmp_var = operator(np.ma.array(aux_cube, mask=mask_cube), axis=0)
        tmp_ncombine = np.ma.sum(np.ma.array(np.ones_like(data_cube),
                                             mask=mask_cube), axis=0)
        with np.errstate(divide='ignore', over='ignore'):
            tmp_var /= tmp_ncombine
    
    if output_var is not None:
        output_var[row_limits[0]:row_limits[1]] = tmp_var
        
    # Produce a proper DQ output, since np.ma doesn't differentiate
    # between different DQ bits (unlike NDData's mask)
    if output_mask is not None:
        tmp_mask = np.where(reduce(np.bitwise_and,
                            [plane for plane in mask_cube])==16, 16, tmp_mask)
        output_mask[row_limits[0]:row_limits[1]] = tmp_mask