import numpy as np
from astropy.modeling import models, projections, Model, Parameter
from astropy.io import fits
from copy import deepcopy

from .utils import merge_objcats
from .lookups import general_parameters as params
from .bookkeeping.log_utilities import get_logger

log = get_logger(__name__)

class CallableWCS(Model):
    """
    Wrapper to make an astropy.WCS object act like an astropy.modeling.Model
    object, including having an inverse.
    """
    inputs = ('x','y')
    outputs = ('x','y')
    direction = Parameter()

    def __init__(self, wcs, direction=1):
        self._wcs = wcs
        super(CallableWCS, self).__init__(direction)
    
    def evaluate(self, x, y, direction):
        return self._wcs.all_pix2world(x, y, 0) if direction>0 \
            else self._wcs.all_world2pix(x, y, 0)
    
    @property
    def inverse(self):
        return CallableWCS(self._wcs, direction=-self.direction)
    
    @property
    def wcs(self):
        return self._wcs
########################## End of CallableWCS Class ##########################

class Transform(object):
    """
    A chained set of astropy Models with the ability to select a subchain.
    Since many transformations will be concatenated pairs of Models (one
    for each ordinate), this also creates a better structure.
    """
    def __init__(self, model=None):
        """Initialize with a single model, list, or empty"""
        self._models = []
        if model:
            self.update(model)
            
    def update(self, model):
        """Append a Model(s) to the end of the current Transform"""
        if model is not None:
            if isinstance(model, list):
                self._models.extend(model)
            else:
                self._models.append(model)
                
    def map(self, start=None, finish=None, reverse=False):
        """
        Create a CompoundModel from the Transform (or a subset thereof)
        that performs a mapping
        
        Inputs: start=index of the model to start from (None=beginning)
                finish=index of the model to finish with (None=end)
                reverse=perform the reverse mapping?
        """
        if not self._models:
            return Identity(2)
    
        if finish is not None:
            models_to_link = self._models[start:finish+1]
        else:
            models_to_link = self._models[start:finish]

        if reverse:
            composite_model = models_to_link[-1].inverse
            for model in models_to_link[-2::-1]:
                composite_model = composite_model | model.inverse
        else:
            composite_model = models_to_link[0]
            for model in models_to_link[1:]:
                composite_model = composite_model | model
                
        return composite_model
            
    def __call__(self, x_in, y_in, start=None, finish=None, reverse=False):
        """
        Perform transformation using the (sub)chain
        
        Parameters
        ----------
        x_in: value or list
            input 'x' values (pixels, RA, whatever)
        y_in: value or list
            input 'y' values (pixels, dec, whatever)
        start: int (or None)
            index of starting link in transformation chain (None means start)
        finish: int (or None)
            index of finishing link (non-pythonic) in chain (None means end)
        reverse: bool
            set to run the transformation backwards?
        """
        return self.map(start, finish, reverse)(x_in, y_in)
    
    def info(self):
        """Print out something vaguely intelligible for debugging purposes"""
        new_line_indent = "\n    "
        msg = ['Model information:']
        for model in self._models:
            msg.append(repr(model))
        return new_line_indent.join(msg)
########################### End of Transform Class ###########################
    
class GeoMapping(object):
    """
    Needed to create a callable object for the geometric_transform. Stores two
    mapping arrays (one for x, one for y) and provides a function that returns
    the value of the array at the location of the tuple passed to that function
    """
    def __init__(self, xmap, ymap):
        self._xmap = xmap
        self._ymap = ymap
        
    def __call__(self, coords):
        y, x = coords
        return (self._ymap[y, x], self._xmap[y, x])
########################## End of GeoMapping Class ###########################
    
def make_tnx_model(forward_model, reverse_model):
    """
    Return an instance of a Pix2Sky_TNX transform and its inverse
    Inputs: forward_model = list of dicts describing forward transformation
            reverse_model = list of dicts describing reverse transformation
            These dicts need to have the following keys:
              model = name of model (only Polynomial2D, Legendre2D,
                                          Chebyshev2D are currently supported)
              parameters = dict with 'x' and 'y' keys listing the parameters
                           for the model for each ordinate
    """
    forward_transform = None
    for direction in [forward_model, reverse_model]:
        for counter, model in enumerate(direction):
            model_type = model['model']
            transform = {}
            for ordinate in ['x','y']:
                # TODO Act reasonably if unacceptable model is passed

                # Get the degree of the model (needed to instantiate)
                # Parameter (coefficient) names are of form cm_n, where m
                # is the degree in x and n is the degree in y
                max_x_power = np.max([int(key[1])
                                      for key in model['parameters'][ordinate]])
                max_y_power = np.max([int(key[3])
                                      for key in model['parameters'][ordinate]])
                if model_type == 'Polynomial2D':
                    t = models.Polynomial2D(degree=max(max_x_power,max_y_power))
                elif model_type in ['Chebyshev2D','Legendre2D']:
                    t = getattr(models, model_type)(x_degree=max_x_power,
                                                    y_degree=max_y_power)
                else:
                    pass
                # Set model parameters
                parameters = model['parameters'][ordinate]
                for param_name, param_value in parameters.items():
                    setattr(t, param_name, param_value)
                transform.update({ordinate: t})
    
            # Construct a model that takes (x,y) as inputs and outputs (x',y')
            # Since both x' and y' need (x,y) as inputs, we must duplicate inputs
            m = models.Mapping((0,1,0,1)) | (transform['x'] & transform['y'])

            if counter == 0:
                direction_model = m
            else:
                direction_model = direction_model + m

        # Create the transform / inverse
        if forward_transform:
            forward_transform.inverse = direction_model
        else:
            forward_transform = direction_model

    return forward_transform.rename('Pix2Pix_TNX')
#-----------------------------------------------------------------------------
def calculate_output_projection(datasets, phu, projection=params.PROJECTION,
                                pixel_scale=params.PIXEL_SCALE, final_pa=None):
    """
    Attach the output transformation from (RA,dec) to pixels and calculate
    output image size
    Inputs: datasets = list of BartData instances
            projection = default projection to use
            pixscale = output plate scale (arcsec/pixel)
    Outputs: output_map = list of models to make the output projection
             output_shape = tuple giving dimensions of output image
    """
    
    # Take the RA and dec of the reference image as the fiducial point
    ra = phu['RA']
    dec = phu['DEC']
    if final_pa is None:
        final_pa = phu['PA']
        log.debug("Output image will have PA={}".format(final_pa))

    # Check that the projection being used is acceptable
    projection = getattr(models, projection)
    assert issubclass(projection, projections.Zenithal), \
        "A Zenithal output projection is required"
        
    # First stages of the output projection, which map from (RA,dec) to a
    # pixel grid
    output_map = [models.RotateCelestial2Native(ra, dec, 180),
                   projection(),
                  (models.Scale(3600/pixel_scale) &
                   models.Scale(3600/pixel_scale)).rename('DegreesToPixels')]
    
    # Final rotation
    if np.abs(final_pa) > 0.01:
        output_map.append(models.Rotation2D(final_pa))

    # Go through all the inputs and find the limits of this partial output
    # projection, using the corners of each array as fiducial points    
    x_out = []
    y_out = []
    for identifier, dataset in datasets.items():
        for extver, array in dataset.items():
            ysize, xsize = array['shape']
            x_corners = [0, xsize-1, xsize-1, 0]
            y_corners = [0, 0, ysize-1, ysize-1]
            wcs_transform = deepcopy(array['wcs'])
            wcs_transform.update(output_map)
            xout, yout = wcs_transform(x_corners, y_corners)
            log.debug('Corners of {}:{} project to:'.format(identifier,extver))
            log.debug(xout)
            log.debug(yout)
            x_out.extend(xout)
            y_out.extend(yout)
    xmin = np.min(x_out)
    xmax = np.max(x_out)
    ymin = np.min(y_out)
    ymax = np.max(y_out)
    
    log.debug("Extrema: x=({},{}) y =({},{})".format(xmin,xmax,ymin,ymax))
    output_map.append((models.Shift(-xmin) & models.Shift(-ymin)).
                      rename('Centering shift'))
    output_shape = (int(np.ceil(ymax-ymin))+1, int(np.ceil(xmax-xmin))+1)
    log.debug("Output transform: "+Transform(output_map).info())
    
    return output_map, output_shape