import numpy as np
from astropy.modeling import models, fitting
from astropy.table import vstack

def measure_bg_from_image(image, sampling=10):
    """
    Measure the background level by Gaussian fitting
    
    Inputs: image = NDData instance
            sampling = 1-in-n sampling of pixels
    Output: background value
    """
    
    if hasattr(image, 'mask'):
        if image.mask is not None:
            bg_data = image.data[image.mask==0].flatten()
        else:
            bg_data = image.data.flatten()
    else:
        # Sent an array, not an NDData instance
        bg_data = image.data.flatten()

    bg_data = bg_data[::sampling]
    bg = np.median(bg_data)
    bg_std = np.std(bg_data)
    binsize = bg_std*0.1
    # Fit from -5 to +1 sigma
    bins = np.arange(bg-5*bg_std, bg+bg_std, binsize)
    histdata, _ = np.histogram(bg_data, bins)
    # bin centers
    x = bins[:-1] + 0.5*(bins[1]-bins[0])
    # Eliminate bins with no data (e.g., if data are quantized)
    x = x[histdata>0]
    histdata = histdata[histdata>0]
    g_init = models.Gaussian1D(amplitude=np.max(histdata),
                               mean=bg, stddev=bg_std)
    fit_g = fitting.LevMarLSQFitter()
    g = fit_g(g_init, x, histdata)
    bg, bg_std = g.mean.value, abs(g.stddev.value)

    return bg
#-----------------------------------------------------------------------------
def merge_objcats(dataset):
    """
    Return a single Table of the merged OBJCATS of the input dataset
    """
    # Can deal with dataset being BartData or part of the controlling dict
    if isinstance(dataset, dict):
        objcats = [array['OBJCAT'] for array in dataset.values()
                   if 'OBJCAT' in array]
    else:
        objcats = [array.meta['OBJCAT'] for array in dataset
                   if 'OBJCAT' in array.meta]
    return vstack(objcats, metadata_conflicts='silent')